#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include "bussim_dhid.h"
#include "netex_line_master.h"
#include "netex_line.h"
#include "PUGIXML/pugixml.hpp"

using namespace std;

DHID_ID::DHID_ID()
{
    this->ORIGINAL_CODE = "undefined";
    this->CountryCode = "undefined";
    this->AdminArea = "undefined";
    this->StopPlace = "undefined";
    this->Level = "undefined";
    this->Quay = "undefined";
    this->StopPosition = "undefined";
    this->OriginalName = "undefined";
    this->is_set = 1;
    this->longitude = 0.0;
    this->latitude = 0.0;
}

DHID_ID::DHID_ID(std::string _ORIGINAL_CODE, std::string _ORIGINALNAME, double _LONGITUDE, double _LATITUDE)
{
    // set the code components
    this->from_string(_ORIGINAL_CODE);

    // set the original name
    this->OriginalName = _ORIGINALNAME;

    // set the geocoordinates
    this->longitude = _LONGITUDE;
    this->latitude = _LATITUDE;

    // indicate this object as set
    this->is_set = 1;
}

DHID_ID::~DHID_ID()
{

}

void DHID_ID::from_string(std::string _STRTOCONVERT)
{
    // this procedure splits a string in the form
    // <CountryCode>:<AdminArea>:<StopPlace>:<Level>:<Quay>:<StopPosition>
    // into the different attributes

    this->ORIGINAL_CODE = _STRTOCONVERT;
    std::string str;

    // ss is an object of stringstream that references the S string.
    std::stringstream ss(_STRTOCONVERT);

    // Use while loop to check the getline() function condition.
    int cntr=0;
    while (getline(ss, str, ':'))
    {
        switch( cntr )
        {
            case 0:
                this->CountryCode = str;
                cntr++;
                break;
            case 1:
                this->AdminArea = str;
                cntr++;
                break;
            case 2:
                this->StopPlace = str;
                cntr++;
                break;
            case 3:
                this->Level = str;
                cntr++;
                break;
            case 4:
                this->Quay = str;
                cntr++;
                break;
            case 5:
                this->StopPosition = str;
                cntr++;
                break;
    }
    }
}

BUSSIM_DHID_StopPlace::BUSSIM_DHID_StopPlace()
{
    // this is the standard constructor
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;

    this->NETEX_Id = "";
}

BUSSIM_DHID_StopPlace::BUSSIM_DHID_StopPlace(class DHID_ID _ID)
{
    // this is an overloaded constructor processing a given ID
    this->ID = _ID;
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;

    this->NETEX_Id = "";
}

BUSSIM_DHID_StopPlace::BUSSIM_DHID_StopPlace(class DHID_ID _ID, std::string _NETEX_Id)
{
    // this is an overloaded constructor processing a given ID
    this->ID = _ID;
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;

    this->NETEX_Id = _NETEX_Id;
}

BUSSIM_DHID_StopPlace::~BUSSIM_DHID_StopPlace()
{
    // this is the standard de-constructor
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;
}

void BUSSIM_DHID_StopPlace::print(void)
{
    std::cout << this->ID.CountryCode << "\t"
                    << this->ID.AdminArea << "\t"
                    << this->ID.StopPlace << "\t"
                    << this->ID.Level << "\t"
                    << this->ID.Quay << "\t"
                    << this->ID.StopPosition << "\t"
                    << this->ID.OriginalName << "\t"
                    << this->ID.longitude << "\t"
                    << this->ID.latitude << "\t"
                    << this->NETEX_Id
                    << std::endl;
}

BUSSIM_DHID_Quay::BUSSIM_DHID_Quay()
{
    // this is the standard constructor
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;

    this->NETEX_Id = "";
    this->SERVED = 0;
}

BUSSIM_DHID_Quay::BUSSIM_DHID_Quay(class DHID_ID _ID)
{
    // this is an overloaded constructor processing a given ID
    this->ID = _ID;
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;

    this->NETEX_Id = "";
    this->SERVED = 0;
}

BUSSIM_DHID_Quay::BUSSIM_DHID_Quay(class DHID_ID _ID, std::string _NETEX_Id)
{
    // this is an overloaded constructor processing a given ID
    this->ID = _ID;
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;

    this->NETEX_Id = _NETEX_Id;
    this->SERVED = 0;
}

void BUSSIM_DHID_Quay::print(void)
{
    std::cout << this->ID.CountryCode << "\t"
                    << this->ID.AdminArea << "\t"
                    << this->ID.StopPlace << "\t"
                    << this->ID.Level << "\t"
                    << this->ID.Quay << "\t"
                    << this->ID.StopPosition << "\t"
                    << this->ID.OriginalName << "\t"
                    << this->ID.longitude << "\t"
                    << this->ID.latitude << "\t"
                    << "served=" << this->SERVED
                    << std::endl;
}

BUSSIM_DHID_Quay::~BUSSIM_DHID_Quay()
{
    // this is the standard de-constructor
    this->ptr2prev = nullptr;
    this->ptr2next = nullptr;
}


BUSSIM_DHID_StopPlaceList::BUSSIM_DHID_StopPlaceList()
{
    // this is the standard constructor
    this->StopPlaces = 0;
    this->ref2first = nullptr;
    this->ref2last = nullptr;
}

void BUSSIM_DHID_StopPlaceList::Append_StopPlace(class BUSSIM_DHID_StopPlace *_StopPlace)
{
    if( this->ref2first != nullptr )
    {
        _StopPlace->ptr2prev = this->ref2last;
        _StopPlace->ptr2next = nullptr;
        this->ref2last->ptr2next = _StopPlace;
        this->ref2last = _StopPlace;
        this->StopPlaces++;
    }
    else
    {
        // nothing stored so far
        this->ref2first = _StopPlace;
        this->ref2last = _StopPlace;
        this->StopPlaces++;
    }
}

void BUSSIM_DHID_StopPlaceList::print(void)
{
    class BUSSIM_DHID_StopPlace *Tmp;
    Tmp = this->ref2first;
    while( Tmp != nullptr )
    {
        Tmp->print();
        Tmp = Tmp->ptr2next;
    }

}

long BUSSIM_DHID_StopPlaceList::get_stopplaces(void)
{
    return( this->StopPlaces );
}

class BUSSIM_DHID_StopPlace *BUSSIM_DHID_StopPlaceList::already_stored_level3(class DHID_ID _ID)
{
    // prepare the return value
    class BUSSIM_DHID_StopPlace *RetPtr, *TmpPtr;
    RetPtr = nullptr;

    // run through the list of stored StopPlaceItems
    TmpPtr = this->ref2first;
    while( TmpPtr != nullptr )
    {
        if( (TmpPtr->ID.CountryCode == _ID.CountryCode)
            &&
            (TmpPtr->ID.AdminArea == _ID.AdminArea)
            &&
            (TmpPtr->ID.StopPlace == _ID.StopPlace)
           )
        {
            // matching item found
            RetPtr = TmpPtr;
            TmpPtr = nullptr;
        }
        else
        {
            // go to next item
            TmpPtr = TmpPtr->ptr2next;
        }
    }

    return(RetPtr);
}

BUSSIM_DHID_StopPlaceList::~BUSSIM_DHID_StopPlaceList()
{
    // this is the standard destructor

    if( this->ref2first != nullptr )
    {
        // some this is stored here
        if( this->ref2first == this->ref2last )
        {
            // exactly one item is stored
            delete this->ref2first;
            this->ref2first = nullptr;
            this->ref2last = nullptr;
            this->StopPlaces--;
        }
        else
        {
            // at least two items are stored
            class BUSSIM_DHID_StopPlace *TmpPtr;
            TmpPtr = this->ref2last;
            while( TmpPtr->ptr2prev != nullptr )
            {
                TmpPtr = TmpPtr->ptr2prev;
                delete TmpPtr->ptr2next;
                TmpPtr->ptr2next = nullptr;
                this->ref2last = TmpPtr;
                this->StopPlaces--;
            }
            if( TmpPtr != nullptr )
            {
                // exactly one item is stored
                delete TmpPtr;
                this->ref2first = nullptr;
                this->ref2last = nullptr;
                this->StopPlaces--;
            }
        }
    }
}

BUSSIM_DHID_QuayList::BUSSIM_DHID_QuayList()
{
    // the standard constructor
    this->Quays = 0;
    this->ref2first = nullptr;
    this->ref2last = nullptr;

}

BUSSIM_DHID_QuayList::~BUSSIM_DHID_QuayList()
{
    // the standard destructor
        if( this->ref2first != nullptr )
    {
        // some this is stored here
        if( this->ref2first == this->ref2last )
        {
            // exactly one item is stored
            delete this->ref2first;
            this->ref2first = nullptr;
            this->ref2last = nullptr;
            this->Quays--;
        }
        else
        {
            // at least two items are stored
            class BUSSIM_DHID_Quay *TmpPtr;
            TmpPtr = this->ref2last;
            while( TmpPtr->ptr2prev != nullptr )
            {
                TmpPtr = TmpPtr->ptr2prev;
                delete TmpPtr->ptr2next;
                TmpPtr->ptr2next = nullptr;
                this->ref2last = TmpPtr;
                this->Quays--;
            }
            if( TmpPtr != nullptr )
            {
                // exactly one item is stored
                delete TmpPtr;
                this->ref2first = nullptr;
                this->ref2last = nullptr;
                this->Quays--;
            }
        }
    }

}

void BUSSIM_DHID_QuayList::Append_Quay(class BUSSIM_DHID_Quay *_Quay)
{
    if( this->ref2first != nullptr )
    {
        _Quay->ptr2prev = this->ref2last;
        _Quay->ptr2next = nullptr;
        this->ref2last->ptr2next = _Quay;
        this->ref2last = _Quay;
        this->Quays++;
    }
    else
    {
        // nothing stored so far
        this->ref2first = _Quay;
        this->ref2last = _Quay;
        this->Quays++;
    }
}

void BUSSIM_DHID_QuayList::print(void)
{
    class BUSSIM_DHID_Quay *Tmp;
    Tmp = this->ref2first;
    while( Tmp != nullptr )
    {
        Tmp->print();
        Tmp = Tmp->ptr2next;
    }

}

long BUSSIM_DHID_QuayList::get_quays(void)
{
    return( this->Quays );
}

class BUSSIM_DHID_Quay *BUSSIM_DHID_QuayList::already_stored_level3(class DHID_ID _ID)
{
    // prepare the return value
    class BUSSIM_DHID_Quay *RetPtr, *TmpPtr;
    RetPtr = nullptr;

    // run through the list of stored StopPlaceItems
    TmpPtr = this->ref2first;
    while( TmpPtr != nullptr )
    {
        if( (TmpPtr->ID.CountryCode == _ID.CountryCode)
            &&
            (TmpPtr->ID.AdminArea == _ID.AdminArea)
            &&
            (TmpPtr->ID.StopPlace == _ID.StopPlace)
           )
        {
            // matching item found
            RetPtr = TmpPtr;
            TmpPtr = nullptr;
        }
        else
        {
            // go to next item
            TmpPtr = TmpPtr->ptr2next;
        }
    }

    return(RetPtr);
}

class BUSSIM_DHID_Quay *BUSSIM_DHID_QuayList::already_stored_netexid(std::string _NETEX_Id)
{
    // prepare the return value
    class BUSSIM_DHID_Quay *RetPtr, *TmpPtr;
    RetPtr = nullptr;

    // run through the list of stored StopPlaceItems
    TmpPtr = this->ref2first;
    while( TmpPtr != nullptr )
    {
        if( (TmpPtr->NETEX_Id == _NETEX_Id)
           )
        {
            // matching item found
            RetPtr = TmpPtr;
            TmpPtr = nullptr;
        }
        else
        {
            // go to next item
            TmpPtr = TmpPtr->ptr2next;
        }
    }

    return(RetPtr);
}

class BUSSIM_DHID_Quay *BUSSIM_DHID_QuayList::already_stored_level5(class DHID_ID _ID)
{
    // prepare the return value
    class BUSSIM_DHID_Quay *RetPtr, *TmpPtr;
    RetPtr = nullptr;

    // run through the list of stored StopPlaceItems
    TmpPtr = this->ref2first;
    while( TmpPtr != nullptr )
    {
        if( (TmpPtr->ID.CountryCode == _ID.CountryCode)
            &&
            (TmpPtr->ID.AdminArea == _ID.AdminArea)
            &&
            (TmpPtr->ID.StopPlace == _ID.StopPlace)
            &&
            (TmpPtr->ID.Level == _ID.Level)
            &&
            (TmpPtr->ID.Quay == _ID.Quay)
           )
        {
            // matching item found
            RetPtr = TmpPtr;
            TmpPtr = nullptr;
        }
        else
        {
            // go to next item
            TmpPtr = TmpPtr->ptr2next;
        }
    }

    return(RetPtr);
}

DHID_LIST::DHID_LIST()
{
    // this is the standard constructor
    this->StopPlaceList = new class BUSSIM_DHID_StopPlaceList();
    this->QuayList = new class BUSSIM_DHID_QuayList();
    this->MasterLIneList = new class NetexLineMaster_List();
    this->LineList = new class NetexLine_List();
    this->ScenarioName = "undefined list name";

    this->FILES = 0;
    this->FILENAME = nullptr;
}

DHID_LIST::~DHID_LIST()
{

}

int DHID_LIST::get_files(void)
{
    return( this->FILES);
}

std::string DHID_LIST::get_file_name_at_list_pos(int _ListPos)
{
    return( this->FILENAME[_ListPos] );
}

void DHID_LIST::import_netexfile_list(std::string _FILE)
{
    // read the toi-XML document into memory
    pugi::xml_document doc;

    std::ifstream stream(_FILE);
    pugi::xml_parse_result result = doc.load(stream);

    if( result )
    {
        std::cout << "netex file list " << _FILE << " opened successfully!" << std::endl;

        // count the number of files to be processed
        this->FILES = 0;
        pugi::xml_node First_FN = doc.root().child("netex_files").child("file");
        for( pugi::xml_node FN=First_FN ; FN ; FN = FN.next_sibling("file"))
        {
            this->FILES++;
        }

        // setup the array of filenames
        std::cout << "we have " << this->FILES << " files to be processed" << std::endl;
        if( this->get_files() > 0 )
        {
            this->FILENAME = new std::string [this->get_files()];
            First_FN = doc.root().child("netex_files").child("file");
            int cntr=0;
            for( pugi::xml_node FN=First_FN ; FN ; FN = FN.next_sibling("file"))
            {
                this->FILENAME[cntr] = FN.attribute("name").as_string();
                cntr++;
            }
        }
        else
        {
            std::cout << "no netex files specified!" << std::endl;
            exit(0);
        }
    }
}

void DHID_LIST::set_name(std::string _ScenarioName)
{
    this->ScenarioName = _ScenarioName;
}

std::string DHID_LIST::get_name(void)
{
    return( this->ScenarioName );
}

void DHID_LIST::print(void)
{
    std::cout << "We have the Scenario with the name " << this->ScenarioName << std::endl;
    std::cout << "The stored stop places ("<<this->StopPlaceList->get_stopplaces()<<" stop places)" << std::endl;
    this->StopPlaceList->print();
    std::cout << "The stored quays ("<<this->QuayList->get_quays()<<" quays)" << std::endl;
    this->QuayList->print();
    std::cout << "The stored line masters ("<<this->MasterLIneList->get_NetexLineMasters()<<" line masters)" << std::endl;
    this->MasterLIneList->print();
    //std::cout << "The stored lines ("<<this->LineList->get_NetexLines()<<" lines)" << std::endl;
    //this->LineList->print();
}

void DHID_LIST::analyze_from_netex_xml_lines(std::string _FILE)
{
    // read the toi-XML document into memory
    pugi::xml_document doc;

    std::ifstream stream(_FILE);
    pugi::xml_parse_result result = doc.load(stream);

    if( result )
    {
        std::cout << "successfully opened netex-xml file with the name " << _FILE << std::endl;

        std::string AUTHORITY = std::string(doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ResourceFrame")
                                            .child("organisations")
                                            .child("Authority").child_value("PublicCode"));

        std::string OPERATOR = std::string(doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ResourceFrame")
                                            .child("organisations")
                                            .child("Operator").child_value("PublicCode"));

        pugi::xml_node FirstLineMaster = doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ServiceFrame")
                                            .child("lines")
                                            .child("Line");

        for( pugi::xml_node LineMaster=FirstLineMaster ; LineMaster ; LineMaster=LineMaster.next_sibling("Line"))
        {
            //std::cout << "we have the LineMaster with" << std::endl;
            //std::cout << AUTHORITY << "\t" << OPERATOR << "\t" << "id:" << LineMaster.attribute("id").as_string() << std::endl;

            class NetexLineMaster *NLM_TMP;
            NLM_TMP = new class NetexLineMaster(LineMaster.attribute("id").as_string(),
                                                AUTHORITY,
                                                OPERATOR,
                                                std::string(LineMaster.child_value("ShortName")));
            this->MasterLIneList->Append_NetexLineMaster(NLM_TMP);
            NLM_TMP = nullptr;
        }

        // save the journeypattern (aka bussim_line-instances)

        pugi::xml_node FirstLine = doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ServiceFrame")
                                            .child("journeyPatterns")
                                            .child("ServiceJourneyPattern");

        std::string ServiceJourneyPatternID;
        std::string NetexLine_ID;
        std::string LineNumberPublic;
        std::string Mode;
        std::string HexColor;
        long ID;
        double RED = 1.0;
        double GREEN = 1.0;
        double BLUE = 1.0;
        for( pugi::xml_node Line=FirstLine ; Line ; Line=Line.next_sibling("ServiceJourneyPattern"))
        {
            //std::cout << "----------------- new line ------------------------ " << std::endl;
            ServiceJourneyPatternID = Line.attribute("id").as_string();
            NetexLine_ID = Line.child("RouteView").child("LineRef").attribute("ref").as_string();
            ID = this->LineList->get_NetexLines();

            // now inspect the line specification to get line details
            pugi::xml_node FirstLineInf = doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ServiceFrame")
                                            .child("lines")
                                            .child("Line");
            for( pugi::xml_node LineInf=FirstLineInf ; LineInf ; LineInf=LineInf.next_sibling("Line"))
            {
                if( LineInf.attribute("id").as_string() == NetexLine_ID )
                {
                    // we found the line
                    Mode = std::string(LineInf.child_value("TransportMode"));
                    HexColor = std::string(LineInf.child("Presentation").child_value("Colour"));
                    if ( HexColor == "" )
                        HexColor = "FFFFFF";
                    int r,g,b;
                    std::sscanf(HexColor.c_str(),"%02x%02x%02x", &r, &g, &b);
                    // std::cout << r << "," << g << "," << b << std::endl;
                    RED = (double)(r) / (double)(255);
                    GREEN = (double)(g) / (double)(255);
                    BLUE = (double)(b) / (double)(255);

                    LineNumberPublic = std::string(LineInf.child_value("PublicCode"));
                }
            }

            // add the new line to LineList if necessary
            class NetexLine *LinePtr;
            LinePtr = this->LineList->already_stored_ServiceJourneyPattern_ID(ServiceJourneyPatternID);
            if( LinePtr != nullptr )
            {
                // already stored
            }
            else
            {
                // not yet stored
                LinePtr = new class NetexLine(NetexLine_ID,ServiceJourneyPatternID,ID,LineNumberPublic,Mode,RED,GREEN,BLUE);
                this->LineList->Append_NetexLine(LinePtr);
                LinePtr = nullptr;
            }

            // and now try to add it to the associated line master
            class NetexLineMaster *Ptr2LineMaster;
            Ptr2LineMaster = this->MasterLIneList->already_stored_netexid(NetexLine_ID);
            if( Ptr2LineMaster != nullptr )
            {
                // attach currently considered servicepattern/line to list of lines assoc with Ptr2LineMaster
                LinePtr = new class NetexLine(NetexLine_ID,ServiceJourneyPatternID,ID,LineNumberPublic,Mode,RED,GREEN,BLUE);
                Ptr2LineMaster->AssocLineList->Append_NetexLine(LinePtr);


                // now inspect and store the service pattern
                // PIS = Point in Sequence
                pugi::xml_node FirstSPIJP = Line.child("pointsInSequence").child("StopPointInJourneyPattern");

                std::string SSPRef = "";
                int SSPRefForBoarding = 0;
                int SSPRefForAlighting = 0;
                int SSPRefChangeOfDesinationDisplay = 0;
                std::string DisplayRef = "undefinded dest ref";
                std::string DisplayTextFront = "undefined front dest";
                std::string DisplayTextSide = "undefined side dest";
                for( pugi::xml_node SPIJP=FirstSPIJP ; SPIJP ; SPIJP=SPIJP.next_sibling("StopPointInJourneyPattern"))
                {
                    SSPRef = SPIJP.child("ScheduledStopPointRef").attribute("ref").as_string();
                    if( (std::string)(SPIJP.child("ForBoarding").child_value())=="true" )
                        SSPRefForBoarding = 1;
                    else
                        SSPRefForBoarding = 0;
                    if( (std::string)(SPIJP.child("ForAlighting").child_value())=="true" )
                        SSPRefForAlighting = 1;
                    else
                        SSPRefForAlighting = 0;
                    if( (std::string)(SPIJP.child("ChangeOfDestinationDisplay").child_value())=="true" )
                        SSPRefChangeOfDesinationDisplay = 1;
                    else
                        SSPRefChangeOfDesinationDisplay = 0;

                    // SSPRefChangeOfDesinationDisplay = 0;
                    DisplayTextFront = "undefined front dest";
                    DisplayTextSide = "undefined site dest";

                    // destination raussuchen
                    DisplayRef = SPIJP.child("DestinationDisplayRef").attribute("ref").as_string();
                    pugi::xml_node FirstDestination = doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ServiceFrame")
                                            .child("destinationDisplays")
                                            .child("DestinationDisplay");
                    for( pugi::xml_node Destination = FirstDestination; Destination; Destination = Destination.next_sibling("DestinationDisplay"))
                    {
                        if( Destination.attribute("id").as_string() == DisplayRef )
                        {
                            DisplayTextFront = Destination.child_value("FrontText");
                            DisplayTextSide = Destination.child_value("SideText");
                        }
                    }
                    //std::cout << "-->" << SSPRef << std::endl;

                    // now have a look into the list of passengerstopassignments (PSA)
                    // and the PSA using the SSP with ref= SSPREF
                    // for this PSA return the DHID from the associated Quay
                    pugi::xml_node FirstPSA = doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ServiceFrame")
                                            .child("stopAssignments")
                                            .child("PassengerStopAssignment");
                    std::string QUAYREF = "";
                    class BUSSIM_DHID_Quay *Ptr2PSAQuay;

                    for( pugi::xml_node PSA=FirstPSA ; PSA ; PSA=PSA.next_sibling("PassengerStopAssignment"))
                    {
                        if( PSA.child("ScheduledStopPointRef").attribute("ref").as_string() == SSPRef )
                        {
                            // std::cout << "SSPREF=" << SSPRef << "\t -> " << PSA.attribute("id").as_string() << "\t -> " << PSA.attribute("order").as_string()<< "\t -> " << PSA.child("QuayRef").attribute("ref").as_string()<< std::endl;
                            QUAYREF = PSA.child("QuayRef").attribute("ref").as_string();
                            Ptr2PSAQuay = this->QuayList->already_stored_netexid(QUAYREF);
                            if( Ptr2PSAQuay != nullptr )
                            {
                                // we have already stored the key (if everything runs regularly)
                                /*
                                std::cout << "STOP at QUAYREF="<<QUAYREF
                                << "\t " << Ptr2PSAQuay->ID.ORIGINAL_CODE
                                << "\t " << Ptr2PSAQuay->ID.OriginalName << std::endl;
                                */

                                // create a new netex line stop object
                                class NetexLineStop *Ptr2Stop;
                                Ptr2Stop = new class NetexLineStop(SSPRef,
                                                                   SSPRefForAlighting,
                                                                   SSPRefForBoarding,
                                                                   SSPRefChangeOfDesinationDisplay,
                                                                   DisplayTextFront,
                                                                   DisplayTextSide,
                                                                   Ptr2PSAQuay->ID,
                                                                   LinePtr);
                                LinePtr->append_netex_line_stop(Ptr2Stop);
                                Ptr2Stop = nullptr;

                                // add it to the line at the end of LinePtr
                            }
                            else
                            {
                                // should be an error
                                std::cout << "STOP at unknown quay!" << std::endl;
                                exit(0);
                            }
                        }
                    }
                }
                LinePtr = nullptr;
            }

        }
    }
}

void DHID_LIST::analyze_from_netex_xml(std::string _FILE)
{
    // read the toi-XML document into memory
    pugi::xml_document doc;

    std::ifstream stream(_FILE);
    pugi::xml_parse_result result = doc.load(stream);

    if( result )
    {
        std::cout << "successfully opened netex-xml file with the name " << _FILE << std::endl;

        pugi::xml_node FirstStopPlace = doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("SiteFrame")
                                            .child("stopPlaces")
                                            .child("StopPlace");

        for( pugi::xml_node StopPlace=FirstStopPlace ; StopPlace ; StopPlace=StopPlace.next_sibling("StopPlace"))
        {
            //std::cout << "ID=" << StopPlace.attribute("id").as_string() << std::endl;
            // first get the 3rd-level-key
            std::string DHID_ORIG_STR = std::string(StopPlace.child("keyList").child("KeyValue").child_value("Value"));
            std::string DHID_ORIG_NAME = std::string(StopPlace.child_value("Name"));
            std::string NETEX_ID = StopPlace.attribute("id").as_string();
            double DHID_ORIG_longitude = atof(StopPlace.child("Centroid").child("Location").child_value("Longitude"));
            double DHID_ORIG_latitude = atof(StopPlace.child("Centroid").child("Location").child_value("Latitude"));
            if( DHID_ORIG_STR == "" )
            {
                std::cout << "stop place " << DHID_ORIG_NAME << " has no DHID! Not added to StopPlaceList" << std::endl;
                //exit(0);
            }
            else
            {
                class DHID_ID tmp = DHID_ID(DHID_ORIG_STR,DHID_ORIG_NAME,DHID_ORIG_longitude,DHID_ORIG_latitude);

                if( this->StopPlaceList->already_stored_level3(tmp) == nullptr )
                {
                    class BUSSIM_DHID_StopPlace *Ptr;
                    Ptr = new class BUSSIM_DHID_StopPlace(tmp,NETEX_ID);
                    this->StopPlaceList->Append_StopPlace(Ptr);
                    Ptr = nullptr;
                }
            }

            // now identify the associated Quays (5th level identifier)
            for( pugi::xml_node Quay=StopPlace.child("quays").child("Quay") ; Quay ; Quay=Quay.next_sibling("Quay"))
            {
                std::string DHID_ORIG_QUAYSTR = std::string(Quay.child("keyList").child("KeyValue").child_value("Value"));
                std::string DHID_ORIG_QUAYNAME = std::string(Quay.child_value("Name"));
                std::string NETEX_QUAY_ID = Quay.attribute("id").as_string();
                double DHID_ORIG_QUAYlongitude = atof(Quay.child("Centroid").child("Location").child_value("Longitude"));
                double DHID_ORIG_QUAYlatitude = atof(Quay.child("Centroid").child("Location").child_value("Latitude"));

                if( DHID_ORIG_QUAYSTR == "" )
                {
                    std::cout << "StopPlace "<< DHID_ORIG_NAME << "("<<NETEX_ID<<")" <<"QUAY " << DHID_ORIG_QUAYNAME << " has no DHID! Not added to QuayList" << std::endl;
                }
                else
                {
                    class DHID_ID tmp = DHID_ID(DHID_ORIG_QUAYSTR,DHID_ORIG_QUAYNAME,DHID_ORIG_QUAYlongitude,DHID_ORIG_QUAYlatitude);

                    if( this->QuayList->already_stored_level5(tmp) == nullptr )
                    {
                        class BUSSIM_DHID_Quay *PtrQ;
                        PtrQ = new class BUSSIM_DHID_Quay(tmp,NETEX_QUAY_ID);
                        this->QuayList->Append_Quay(PtrQ);
                        PtrQ = nullptr;
                    }
                }
            }
        }

        // we continue to read now the <PassengerStopAssignment> - instances
        // we check the associated quay and mark it as "needed" (served)

        pugi::xml_node FirstPSA = doc.child("PublicationDelivery")
                                            .child("dataObjects")
                                            .child("CompositeFrame")
                                            .child("frames")
                                            .child("ServiceFrame")
                                            .child("stopAssignments")
                                            .child("PassengerStopAssignment");

        class BUSSIM_DHID_Quay *Ptr2Quay;
        for( pugi::xml_node PSA=FirstPSA ; PSA ; PSA=PSA.next_sibling("PassengerStopAssignment"))
        {
            //std::cout << "PassengerStopAssignment " << PSA.attribute("id").as_string() << std::endl;
            Ptr2Quay = this->QuayList->already_stored_netexid(PSA.child("QuayRef").attribute("ref").as_string());
            //std::cout << PSA.child("QuayRef").attribute("ref").as_string() << std::endl;
            if( Ptr2Quay != nullptr )
            {
                Ptr2Quay->SERVED++;
            }
            Ptr2Quay = nullptr;
        }
    }
}

void DHID_LIST::export_line_list_to_xml(void)
{

    // Generate new XML document within memory
    pugi::xml_document doc;
    // Alternatively store as shared pointer if tree shall be used for longer
    // time or multiple client calls:
    // std::shared_ptr<pugi::xml_document> spDoc = std::make_shared<pugi::xml_document>();
    // Generate XML declaration
    auto declarationNode = doc.append_child(pugi::node_declaration);
    declarationNode.append_attribute("version")    = "1.0";
    declarationNode.append_attribute("encoding")   = "ISO-8859-1";
    declarationNode.append_attribute("standalone") = "yes";
    // A valid XML doc must contain a single root node of any name
    auto root = doc.append_child("dhid_lines");

    class NetexLineMaster *Ptr2MasterLine;
    Ptr2MasterLine = this->MasterLIneList->ref2first;
    class NetexLine *Ptr2Line;
    Ptr2Line = nullptr;
    while( Ptr2MasterLine != nullptr )
    {
        // write the line master information
        pugi::xml_node LiMa = root.append_child("LineMaster");
        LiMa.append_attribute("NetexID") = Ptr2MasterLine->Netex_ID.c_str();
        LiMa.append_attribute("Authority") = Ptr2MasterLine->Authority.c_str();
        LiMa.append_attribute("Operator") = Ptr2MasterLine->Operator.c_str();
        LiMa.append_attribute("ShortName") = Ptr2MasterLine->ShortName.c_str();
        LiMa.append_attribute("Include") = "1";

        // attach the line information associated with this master
        Ptr2Line = Ptr2MasterLine->AssocLineList->ref2first;
        while( Ptr2Line != nullptr )
        {
            pugi::xml_node Li = LiMa.append_child("Line");
            //Li.append_attribute("NetexLineID") = Ptr2Line->NetexLine_ID.c_str();
            Li.append_attribute("ID") = Ptr2Line->ID;
            Li.append_attribute("ServiceJourneyPatternID") = Ptr2Line->ServiceJourneyPattern_ID.c_str();
            //Li.append_attribute("LineStops") = Ptr2Line->LineStops;
            Li.append_attribute("LineNumberPublic") = Ptr2Line->LineNumberPublic.c_str();
            Li.append_attribute("Mode") = Ptr2Line->Mode.c_str();
            Li.append_attribute("RED") = Ptr2Line->RED;
            Li.append_attribute("GREEN") = Ptr2Line->GREEN;
            Li.append_attribute("BLUE") = Ptr2Line->BLUE;
            Li.append_attribute("Include") = "1";

            // insert the dhid-stop-sequence of this line
            class NetexLineStop *Ptr2Stop;
            Ptr2Stop = Ptr2Line->FirstStop;
            while( Ptr2Stop != nullptr )
            {
                pugi::xml_node Stop = Li.append_child("dhid_object");
                Stop.append_attribute("StopPointInJourneyPattern") = Ptr2Stop->StopPointInJourneyPattern.c_str();
                Stop.append_attribute("ForAlighting") = Ptr2Stop->ForAlighting;
                Stop.append_attribute("ForBoarding") = Ptr2Stop->ForBoarding;
                Stop.append_attribute("ChangeOfDestinationDisplay") = Ptr2Stop->ChangeOfDestinationDisplay;
                Stop.append_attribute("DestinationDisplayTextFront") = Ptr2Stop->DestinationDisplayTextFront.c_str();
                Stop.append_attribute("DestinationDisplayTextSide") = Ptr2Stop->DestinationDisplayTextSide.c_str();
                Stop.append_attribute("DHID") = Ptr2Stop->DHID.ORIGINAL_CODE.c_str();
                Ptr2Stop = Ptr2Stop->next;
            }
            Ptr2Line = Ptr2Line->ptr2next;
        }
        pugi::xml_node Li = LiMa.append_child("Line");

        Ptr2MasterLine = Ptr2MasterLine->ptr2next;
    }
    // save the xml-tree into a file
    bool saveSucceeded = doc.save_file((this->get_name() + "_servicepatterns.xml").c_str(), PUGIXML_TEXT("  "),pugi::format_save_file_text);

}

void DHID_LIST::export_quay_list_to_xml(void)
{
    // Generate new XML document within memory
    pugi::xml_document doc;
    // Alternatively store as shared pointer if tree shall be used for longer
    // time or multiple client calls:
    // std::shared_ptr<pugi::xml_document> spDoc = std::make_shared<pugi::xml_document>();
    // Generate XML declaration
    auto declarationNode = doc.append_child(pugi::node_declaration);
    declarationNode.append_attribute("version")    = "1.0";
    declarationNode.append_attribute("encoding")   = "ISO-8859-1";
    declarationNode.append_attribute("standalone") = "yes";
    // A valid XML doc must contain a single root node of any name
    auto root = doc.append_child("dhid_quays");

    // produce the nodes
    class BUSSIM_DHID_Quay *QuayPtr;
    QuayPtr = this->QuayList->ref2first;

    class BUSSIM_DHID_StopPlace *SPPtr;

    while( QuayPtr != nullptr )
    {
        if( QuayPtr->SERVED > 0 )
        {


        pugi::xml_node QUAY = root.append_child("quay");
        QUAY.append_attribute("ORIGINAL_CODE") = QuayPtr->ID.ORIGINAL_CODE.c_str();
        QUAY.append_attribute("CountryCode") = QuayPtr->ID.CountryCode.c_str();
        QUAY.append_attribute("AdminArea") = QuayPtr->ID.AdminArea.c_str();
        QUAY.append_attribute("StopPlace") = QuayPtr->ID.StopPlace.c_str();
        QUAY.append_attribute("Level") = QuayPtr->ID.Level.c_str();
        QUAY.append_attribute("Quay") = QuayPtr->ID.Quay.c_str();
        QUAY.append_attribute("StopPosition") = QuayPtr->ID.StopPosition.c_str();
        QUAY.append_attribute("OriginalName") = QuayPtr->ID.OriginalName.c_str();
        QUAY.append_attribute("is_set") = QuayPtr->ID.is_set;
        QUAY.append_attribute("longitude") = QuayPtr->ID.longitude;
        QUAY.append_attribute("latitude") = QuayPtr->ID.latitude;
        QUAY.append_attribute("NETEX_Id") = QuayPtr->NETEX_Id.c_str();
        QUAY.append_attribute("SERVED") = QuayPtr->SERVED;

        // find the name of the associated stopplace
        SPPtr = this->StopPlaceList->already_stored_level3(QuayPtr->ID);
        if( SPPtr != nullptr )
        {
            QUAY.append_attribute("AssocStopPlaceName") = SPPtr->ID.OriginalName.c_str();
            QUAY.append_attribute("AssocStopPlaceCode") = SPPtr->ID.ORIGINAL_CODE.c_str();
            QUAY.append_attribute("AssocStopNetexId") = SPPtr->NETEX_Id.c_str();
        }
        else
        {
            QUAY.append_attribute("AssocStopPlaceName") = "";
            QUAY.append_attribute("AssocStopPlaceCode") = "";
            QUAY.append_attribute("AssocStopNetexId") = "";
        }
        }
        QuayPtr = QuayPtr->ptr2next;
    }
    // save the xml-tree into a file
    bool saveSucceeded = doc.save_file((this->get_name() + "_quays.xml").c_str(), PUGIXML_TEXT("  "),pugi::format_save_file_text);

}

void DHID_LIST::import_from_xml(std::string _FILE)
{
    // read the toi-XML document into memory
    pugi::xml_document doc;

    std::ifstream stream(_FILE);
    pugi::xml_parse_result result = doc.load(stream);

    if( result )
    {
        std::cout << "successfully opened the DHID quay list file with the name " << _FILE << std::endl;

        std::string DHID_CODE;
        std::string DHID_NAME;
        std::string DHID_ASSOCSTOPPLACENAME;
        double DHID_LONGITUDE;
        double DHID_LATITUDE;

        for(pugi::xml_node QUAY = doc.child("dhid_quays").child("quay") ; QUAY ; QUAY=QUAY.next_sibling("quay") )
        {
            DHID_CODE = std::string(QUAY.attribute("ORIGINAL_CODE").as_string());
            DHID_NAME = std::string(QUAY.attribute("OriginalName").as_string());
            DHID_ASSOCSTOPPLACENAME = std::string(QUAY.attribute("AssocStopPlaceName").as_string());
            DHID_LONGITUDE = QUAY.attribute("longitude").as_double();
            DHID_LATITUDE = QUAY.attribute("latitude").as_double();


            class DHID_ID Tmp = DHID_ID(DHID_CODE,DHID_NAME,DHID_LONGITUDE,DHID_LATITUDE);
            if( this->QuayList->already_stored_level5(Tmp) == nullptr )
            {
                class BUSSIM_DHID_Quay *Ptr;
                Ptr = new class BUSSIM_DHID_Quay(Tmp);
                this->QuayList->Append_Quay(Ptr);
                Ptr = nullptr;
            }
            class DHID_ID TmpB = DHID_ID(DHID_CODE,DHID_ASSOCSTOPPLACENAME,DHID_LONGITUDE,DHID_LATITUDE);
            TmpB.Level = "";
            TmpB.Quay = "";
            TmpB.StopPosition = "";
            if( this->StopPlaceList->already_stored_level3(TmpB) == nullptr )
            {
                class BUSSIM_DHID_StopPlace *SPPtr;
                SPPtr = new class BUSSIM_DHID_StopPlace(TmpB);
                this->StopPlaceList->Append_StopPlace(SPPtr);
                SPPtr = nullptr;
            }
        }
    }
    else
    {
        std::cout << "unable to open the DHID quay list file with the name " << _FILE << std::endl;
        exit(0);
    }
}
